/**
 * \file SAM4S_EK_C_Ex.c
 *
 * Created: 8/8/2014 9:27:33 AM
 * Author: Owais
 */ 

#include <stdarg.h>

#include "sam4.h"
#include "printf_stdarg.h"
#include "HSMCI.h"
#include "sd_mmc_protocol.h"

#include "FreeRTOS.h"
#include "task.h"
#include "fat_sl.h"
#include "api_mdriver.h"

/*! \def fsSD_BUFFER_SIZE 
    \brief The number of bytes read/written to the example files at a time. */
#define fsSD_BUFFER_SIZE 				200

/*! \def fsMAX_FILE_NAME_LEN
    \brief 8.3 format, plus null terminator. */
#define fsMAX_FILE_NAME_LEN				13

/*! \def fsPUTC_FILE_SIZE
    \brief The number of bytes written to the file that uses f_putc() and f_getc(). */
#define fsPUTC_FILE_SIZE				100

/*! A buffer used to both create content to write to disk, and read content back
from a disk.  Note there is no mutual exclusion on this buffer. */
static char cSDBuffer[ fsSD_BUFFER_SIZE ];

// Names of directories that are created.
static const char *pcRoot = "/", *pcDirectory1 = "SUB1", *pcDirectory2 = "SUB2", *pcFullPath = "/SUB1/SUB2";

/**
 * \brief A function which uses FreeRTOS+FAT-SL APIs
 * 
 * This function creates and verifies different files on the volume, demonstrating the use of
 * various different API functions.
 */
void vCreateAndVerifySampleFiles( void );

/**
 * \brief A function which creates demo files
 * 
 * This function creates a set of example files in the root directory of the volume using
 * f_write().
 */
static void prvCreateDemoFilesUsing_f_write( void );

/**
 * \brief A function which verifies demo files
 * 
 *  This function uses f_read() to read back and verify the files that were created by
 * prvCreateDemoFilesUsing_f_write().
 */
static void prvVerifyDemoFileUsing_f_read( void );

/**
 * \brief A function which creates demo files
 * 
 *  This function creates an example file in a sub-directory using f_putc().
 */
static void prvCreateDemoFileUsing_f_putc( void );

/**
 * \brief A function which verifies demo files
 * 
 * This function uses f_getc() to read back and verify the file that was created by
 * prvCreateDemoFileUsing_f_putc().
 */
static void prvVerifyDemoFileUsing_f_getc( void );

/**
 * \brief FreeRTOS hook (or callback) functions that are defined in this file.
 */
void vApplicationMallocFailedHook ( void );
void vApplicationIdleHook ( void );
void vApplicationStackOverflowHook ( xTaskHandle pxTask, signed char *pcTaskName );
void vApplicationTickHook ( void );

/**
 *  \brief FreeRTOS task declaration
 *  
 *  Task declaration used for
 *  testing SD card. */
void SD_Test ( void* pvParameters );

/**
 *  \brief FreeRTOS task declaration
 *  
 *  Task declaration used for
 *  generating a blinking of
 *  a LED. */
void Blink ( void* pvParameters );

/**
 *  \brief A function which initializes UART 
 *  
 *  This is a method used to
 *  configure the UART0 
 *  interface of SAM4S. */
void Config_UART0 ( void )
{
	// Set peripheral A on PA9.
	PIOA->PIO_ABCDSR[0] = ~PIO_ABCDSR_P9;
	PIOA->PIO_ABCDSR[1] = ~PIO_ABCDSR_P9;
	
	// Disable PIO controller for PA9.
	PIOA->PIO_PDR = PIO_PDR_P9;
	
	// Set peripheral A on PA10.
	PIOA->PIO_ABCDSR[0] = ~PIO_ABCDSR_P10;
	PIOA->PIO_ABCDSR[1] = ~PIO_ABCDSR_P10;
	
	// Disable PIO controller for PA10.
	PIOA->PIO_PDR = PIO_PDR_P10;
	
	// Enable peripheral clock for UART0.
	PMC->PMC_PCER0 = PMC_PCER0_PID8;

	// Reset and disable receiver & transmitter.
	UART0->UART_CR = UART_CR_RSTRX | UART_CR_RSTTX | UART_CR_RXDIS | UART_CR_TXDIS;

	/** Configure baud rate
	 *  Asynchronous, no oversampling. */
	UART0->UART_BRGR = 65;
	
	// Configure mode.
	UART0->UART_MR = UART_MR_PAR_NO;

	// Disable PDC channel.
	UART0->UART_PTCR = UART_PTCR_RXTDIS | UART_PTCR_TXTDIS;

	// Enable receiver and transmitter.
	UART0->UART_CR = UART_CR_RXEN | UART_CR_TXEN;
	
	// Configure UART in normal mode.
	UART0->UART_MR = UART_MR_CHMODE_NORMAL;
}

/** 
 *  \brief A function which creates a delay
 *  
 *  This is a method used to
 *  generate a delay by 
 *  executing a for loop for
 *  a certain number of iteration. */
void WaitSomeTime ( void )
{
	volatile uint32_t i = 0;
	
	while (i < 1000000)
	{
		i++;
	}
}

/**
 * \brief Application entry point.
 *
 * \return Unused (ANSI-C compatibility).
 */
int main(void)
{	
	// Initialize the SAM system.
    SystemInit();
	// Configure the UART0 interface.
	Config_UART0();
	// Configure the HSMCI interface.
	hsmci_init();
	// Initialize the SD card and format if needed.
	vCreateAndVerifySampleFiles();

	// Create tasks.
	xTaskCreate( SD_Test,	(const char*)"Task 1", 500, NULL, 2, NULL );
	xTaskCreate( Blink,	(const char*)"Task 2", 500, NULL, 1, NULL );
	
	// Start scheduler.
	vTaskStartScheduler();

	for (;;);
	return 0;
}

/** 
 *  \brief A function used for blinking
 *  
 *  This is a task used to
 *  generate a short Blink 
 *  on an external LED. */
void Blink (void* pvParameters)
{
	// Enable output.
	REG_PIOA_OER = 0x100000;
	// Output write enable.
	REG_PIOA_OWER = 0x100000;
	// Declare a variable for timekeeping.
	portTickType xLastWakeTime;
	// Get tick counts of the task.
	xLastWakeTime = xTaskGetTickCount();
	for( ;; )
	{
		// Set and reset output.
		REG_PIOA_ODSR ^= 0x100000;
		// Insert a task delay.
		vTaskDelayUntil( &xLastWakeTime, 500 );
		//vTaskDelay(1000);
	}
}

/** 
 *  \brief A function which tests SD card
 *  
 *  This is a task used to write
 *  a string to the SD card 
 *  using the FreeRTOS FAT FS. */
void SD_Test (void* pvParameters)
{
	REG_PIOA_OER = 0x80000;
	REG_PIOA_OWER = 0x80000;

	char cFileName[ fsMAX_FILE_NAME_LEN ];
	F_FILE *pxFile;
	char *pcString = "This is a test string generated to calculate and evaluate the time required to write data on a SD card using the HSMCI interface of SAM4S";
	// Generate a file name.
	sprintf( cFileName, "TestFile.txt");
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	for( ;; )
	{
		REG_PIOA_ODSR ^= 0x80000;
		taskENTER_CRITICAL();
		// Obtain the current working directory and print out the file name and	the	directory into which the file is being written.
		f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );
		// Open the file TestFile.txt for writing.
		pxFile = f_open( cFileName, "w" );
		if( pxFile != NULL )
		{
			// Write characters to the opened file.
			if( f_write( pcString, 1, 137, pxFile ) == 137)
			{
				// All bytes were written to the file.
				//printf("Content written to file\n\r");
			}
			//printf("File created and saved\n\r");
			// Close the file again.
			f_close( pxFile );
		}
		taskEXIT_CRITICAL();
		vTaskDelayUntil( &xLastWakeTime, 1000 );
	}
}

/** 
 *  /brief A function which uses FreeRTOS+FAT-SL APIs
 *  
 *  This is a method used to
 *  initialize the SD card.
 *  It formats the SD card using
 *  FreeRTOS_FAT FS. */
void vCreateAndVerifySampleFiles( void )
{
	unsigned char ucStatus;
	printf("Please insert an SD card in slot\n\r");
	while((REG_PIOA_PDSR & PIN_CDMCI) != 0);
	printf("Card inserted\n\r");
	WaitSomeTime();	
	// First create the volume.
	ucStatus = f_initvolume( sd_init );

	// It is expected that the volume is not formatted.
	if( ucStatus == F_ERR_NOTFORMATTED )
	{
		// Format the created volume with FAT32 FS.
		ucStatus = f_format( F_FAT32_MEDIA );
	}

	if( ucStatus == F_NO_ERROR )
	{
		// Create a set of files using f_write().
		//prvCreateDemoFilesUsing_f_write();

		// Read back and verify the files that were created using f_write().
		//prvVerifyDemoFileUsing_f_read();

		// Create sub directories two deep then create a file using putc.
		//prvCreateDemoFileUsing_f_putc();

		// Read back and verify the file created by prvCreateDemoFileUsing_f_putc().
		//prvVerifyDemoFileUsing_f_getc();
		printf("SD card initialized\n\r");
	}
}

/** 
 *  \brief A function which creates demo files 
 *  
 *  This is a method used to
 *  create demo file on SD card 
 *  using the f_write API provided by
 *  FreeRTOS FAT FS. */
static void prvCreateDemoFilesUsing_f_write( void )
{
portBASE_TYPE xFileNumber, xWriteNumber;
char cFileName[ fsMAX_FILE_NAME_LEN ];
const portBASE_TYPE xMaxFiles = 5;
long lItemsWritten;
F_FILE *pxFile;
	/** Create xMaxFiles files.  Each created file will be
	 *  ( xFileNumber * fsSD_BUFFER_SIZE ) bytes in length, and filled
	 *  with a different repeating character. */
	for( xFileNumber = 1; xFileNumber <= xMaxFiles; xFileNumber++ )
	{
		// Generate a file name.
		sprintf( cFileName, "TestF%03d.txt", ( int ) xFileNumber );

		// Obtain the current working directory and print out the file name and the directory into which the file is being written.
		f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );

		// Open the file, creating the file if it does not already exist.
		pxFile = f_open( cFileName, "w" );
		configASSERT( pxFile );

		/** Fill the SD buffer with data that will be written to the file.  This
		 *  is just a repeating ascii character that indicates the file number. */
		//memset( cSDBuffer, ( int ) ( '0' + xFileNumber ), fsSD_BUFFER_SIZE );
		
		/** Write the SD buffer to the opened file a number of times.
		 *  The number of times the SD buffer is written to the file
		 *  depends on the file number, so the length of each created file will be different. */
		for( xWriteNumber = 0; xWriteNumber < xFileNumber; xWriteNumber++ )
		{
			lItemsWritten = f_write( cSDBuffer, fsSD_BUFFER_SIZE, 1, pxFile );
			configASSERT( lItemsWritten == 1 );
		}

		// Close the file so another file can be created.
		f_close( pxFile );
	}
}

/** 
 *  \brief A function which verifies demo files
 *  
 *  This is a method used to verify
 *  the demo file created on SD card,
 *  using the f_write API, by f_read API */
static void prvVerifyDemoFileUsing_f_read( void )
{
portBASE_TYPE xFileNumber, xReadNumber;
char cFileName[ fsMAX_FILE_NAME_LEN ];
const portBASE_TYPE xMaxFiles = 5;
long lItemsRead, lChar;
F_FILE *pxFile;

	// Read back the files that were created by prvCreateDemoFilesUsing_f_write().
	for( xFileNumber = 1; xFileNumber <= xMaxFiles; xFileNumber++ )
	{
		// Generate the file name.
		sprintf( cFileName, "root%03d.txt", ( int ) xFileNumber );

		/** Obtain the current working directory and print out the file name and
		 *  the	directory from which the file is being read. */
		f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );

		// Open the file for reading.
		pxFile = f_open( cFileName, "r" );
		configASSERT( pxFile );

		/** Read the file into the SD buffer, checking the file contents are as
		 *  expected.  The size of the file depends on the file number. */
		for( xReadNumber = 0; xReadNumber < xFileNumber; xReadNumber++ )
		{
			// Start with the SD buffer clear.
			memset( cSDBuffer, 0x00, fsSD_BUFFER_SIZE );

			lItemsRead = f_read( cSDBuffer, fsSD_BUFFER_SIZE, 1, pxFile );
			configASSERT( lItemsRead == 1 );

			/** Check the SD buffer is filled with the expected data.  Each
			 *  file contains a different repeating ascii character that indicates
			 *  the number of the file. */
			for( lChar = 0; lChar < fsSD_BUFFER_SIZE; lChar++ )
			{
				configASSERT( cSDBuffer[ lChar ] == ( '0' + ( char ) xFileNumber ) );
			}
		}

		// Close the file.
		f_close( pxFile );
	}
}

/** 
 *  \brief A function which creates demo files
 *  
 *  This is a method used to
 *  create demo file on SD card 
 *  using the f_putc API provided by
 *  FreeRTOS FAT FS. */
static void prvCreateDemoFileUsing_f_putc( void )
{
unsigned char ucReturn;
int iByte, iReturned;
F_FILE *pxFile;
char cFileName[ fsMAX_FILE_NAME_LEN ];

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );

	// Create a sub directory.
	ucReturn = f_mkdir( pcDirectory1 );
	configASSERT( ucReturn == F_NO_ERROR );

	// Move into the created sub-directory.
	ucReturn = f_chdir( pcDirectory1 );
	configASSERT( ucReturn == F_NO_ERROR );

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );

	// Create a subdirectory in the new directory.
	ucReturn = f_mkdir( pcDirectory2 );
	configASSERT( ucReturn == F_NO_ERROR );

	// Move into the directory just created - now two directories down from the root.
	ucReturn = f_chdir( pcDirectory2 );
	configASSERT( ucReturn == F_NO_ERROR );

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );
	configASSERT( strcmp( cSDBuffer, pcFullPath ) == 0 );

	// Generate the file name.
	sprintf( cFileName, "%s.txt", pcDirectory2 );

	// Print out the file name and the directory into which the file is being written.
	pxFile = f_open( cFileName, "w" );

	// Create a file 1 byte at a time.  The file is filled with incrementing ascii characters starting from '0'.
	for( iByte = 0; iByte < fsPUTC_FILE_SIZE; iByte++ )
	{
		iReturned = f_putc( ( ( int ) '0' + iByte ), pxFile );
		configASSERT( iReturned ==  ( ( int ) '0' + iByte ) );
	}

	// Finished so close the file.
	f_close( pxFile );

	// Move back to the root directory.
	ucReturn = f_chdir( "../.." );
	configASSERT( ucReturn == F_NO_ERROR );

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );
	configASSERT( strcmp( cSDBuffer, pcRoot ) == 0 );
}

/** 
 *  \brief A function which verifies demo files
 *  
 *  This is a method used to verify
 *  the demo file created on SD card,
 *  using the f_putc API, by f_getc API */
static void prvVerifyDemoFileUsing_f_getc( void )
{
unsigned char ucReturn;
int iByte, iReturned;
F_FILE *pxFile;
char cFileName[ fsMAX_FILE_NAME_LEN ];

	// Move into the directory in which the file was created.
	ucReturn = f_chdir( pcFullPath );
	configASSERT( ucReturn == F_NO_ERROR );

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );
	configASSERT( strcmp( cSDBuffer, pcFullPath ) == 0 );

	// Generate the file name.
	sprintf( cFileName, "%s.txt", pcDirectory2 );

	// This time the file is opened for reading.
	pxFile = f_open( cFileName, "r" );

	// Read the file 1 byte at a time.
	for( iByte = 0; iByte < fsPUTC_FILE_SIZE; iByte++ )
	{
		iReturned = f_getc( pxFile );
		configASSERT( iReturned ==  ( ( int ) '0' + iByte ) );
	}

	// Finished so close the file.
	f_close( pxFile );

	// Move back to the root directory.
	ucReturn = f_chdir( "../.." );
	configASSERT( ucReturn == F_NO_ERROR );

	// Obtain and print out the working directory.
	f_getcwd( cSDBuffer, fsSD_BUFFER_SIZE );
}
/**-----------------------------------------------------------*/

void vApplicationMallocFailedHook(void)
{
	/** 
	 *  vApplicationMallocFailedHook() will only be called if
	 *  configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
	 *  function that will get called if a call to pvPortMalloc() fails.
	 *  pvPortMalloc() is called internally by the kernel whenever a task, queue,
	 *  timer or semaphore is created.  It is also called by various parts of the
	 *  demo application.  If heap_1.c or heap_2.c are used, then the size of the
	 *  heap available to pvPortMalloc() is defined by configTOTAL_HEAP_SIZE in
	 *  FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
	 *  to query the size of free heap space that remains (although it does not
	 *  provide information on how the remaining heap might be fragmented). */
	taskDISABLE_INTERRUPTS();
	for (;;) {
	}
}
/**-----------------------------------------------------------*/

void vApplicationIdleHook(void)
{
	/** 
	 *  vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
	 *  to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
	 *  task.  It is essential that code added to this hook function never attempts
	 *  to block in any way (for example, call xQueueReceive() with a block time
	 *  specified, or call vTaskDelay()).  If the application makes use of the
	 *  vTaskDelete() API function (as this demo application does) then it is also
	 *  important that vApplicationIdleHook() is permitted to return to its calling
	 *  function, because it is the responsibility of the idle task to clean up
	 *  memory allocated by the kernel to any task that has since been deleted. */
}
/**-----------------------------------------------------------*/

void vApplicationStackOverflowHook(xTaskHandle pxTask, signed char *pcTaskName)
{
	(void) pcTaskName;
	(void) pxTask;

	/** 
	 *  Run time stack overflow checking is performed if
	 *  configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
	 *  function is called if a stack overflow is detected. */
	taskDISABLE_INTERRUPTS();
	for (;;) {
	}
}
/**-----------------------------------------------------------*/

void vApplicationTickHook(void)
{
	/** 
	 *  This function will be called by each tick interrupt if
	 *  configUSE_TICK_HOOK is set to 1 in FreeRTOSConfig.h.  User code can be
	 *  added here, but the tick hook is called from an interrupt context, so
	 *  code must not attempt to block, and only the interrupt safe FreeRTOS API
	 *  functions can be used (those that end in FromISR()). */
}

