/*
 * HSMCI.h
 *
 * Created: 9/1/2014 1:31:32 PM
 *  Author: Owais
 */ 


#ifndef HSMCI_H_
#define HSMCI_H_

#include "sam.h"

//This SD MMC stack supports only the high voltage
#define SD_MMC_VOLTAGE_SUPPORT (OCR_VDD_27_28 | OCR_VDD_28_29 | OCR_VDD_29_30 | OCR_VDD_30_31 | OCR_VDD_31_32 | OCR_VDD_32_33)
//This SD MMC stack uses the maximum block size authorized (512 bytes)
#define SD_MMC_BLOCK_SIZE          512
//Definition of HSMCI hardware interface
#define PINS_MCI   (PIO_PA26 | PIO_PA27 | PIO_PA28 | PIO_PA29 | PIO_PA30 | PIO_PA31)
//Dard detect pin - software should check it before call sdInit();
#define PIN_CDMCI PIO_PA6

#define OK				1
#define FAIL			0
#define SD_READ_OK		0
#define SD_WRITE_OK		0
#define SD_READ_FAIL	1
#define SD_WRITE_FAIL	1

/**
 * \defgroup sam_drivers_hsmci High Speed MultiMedia Card Interface (HSMCI)
 *
 * This driver interfaces the HSMCI module.
 * It will add functions for SD/MMC card reading, writing and management.
 *
 * @{
 */

/** \brief Initializes the low level driver
 *
 * This enable the clock required and the hardware interface.
 */
uint32_t hsmci_init(void);

void hsmci_reset(void);
void hsmci_set_speed(uint32_t speed, uint32_t mck);
uint32_t hsmci_wait_busy(void);
uint32_t hsmci_send_cmd_execute(uint32_t cmdr, uint32_t cmd, uint32_t arg);

/** \brief Return the maximum bus width of a slot
 *
 * \param slot     Selected slot
 *
 * \return 1, 4 or 8 lines.
 */
uint8_t hsmci_get_bus_width(uint8_t slot);

/** \brief Return the high speed capability of the driver
 *
 * \return OK, if the high speed is supported
 */
uint32_t hsmci_is_high_speed_capable(void);

/**
 * \brief Select a slot and initialize it
 *
 * \param slot       Selected slot
 * \param clock      Maximum clock to use (Hz)
 * \param bus_width  Bus width to use (1, 4 or 8)
 * \param high_speed OK, to enable high speed mode
 */
void hsmci_select_device(uint8_t slot, uint32_t clock, uint8_t bus_width, uint32_t high_speed);

/**
 * \brief Deselect a slot
 *
 * \param slot       Selected slot
 */
void hsmci_deselect_device(uint8_t slot);

/** \brief Send 74 clock cycles on the line of selected slot
 * Note: It is required after card plug and before card install.
 */
void hsmci_send_clock(void);

/** \brief Send a command on the selected slot
 *
 * \param cmd        Command definition
 * \param arg        Argument of the command
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_send_cmd(uint32_t cmd, uint32_t arg);

/** \brief Return the 32 bits response of the last command
 *
 * \return 32 bits response
 */
uint32_t hsmci_get_response(void);

/** \brief Return the 128 bits response of the last command
 *
 * \param response   Pointer on the array to fill with the 128 bits response
 */
void hsmci_get_response_128(uint8_t* response);

/** \brief Send an ADTC command on the selected slot
 * An ADTC (Addressed Data Transfer Commands) command is used
 * for read/write access.
 *
 * \param cmd          Command definition
 * \param arg          Argument of the command
 * \param block_size   Block size used for the transfer
 * \param nb_block     Total number of block for this transfer
 * \param access_block if OK, the x_read_blocks() and x_write_blocks()
 * functions must be used after this function.
 * If FAIL, the mci_read_word() and mci_write_word()
 * functions must be used after this function.
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_adtc_start(uint32_t cmd, uint32_t arg, uint16_t block_size,
		uint16_t nb_block, uint32_t access_block);

/** \brief Send a command to stop an ADTC command on the selected slot
 *
 * \param cmd        Command definition
 * \param arg        Argument of the command
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_adtc_stop(uint32_t cmd, uint32_t arg);

/** \brief Read a word on the line
 *
 * \param value  Pointer on a word to fill
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_read_word(uint32_t* value);

/** \brief Write a word on the line
 *
 * \param value  Word to send
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_write_word(uint32_t value);

/** \brief Start a read blocks transfer on the line
 * Note: The driver will use the DMA available to speed up the transfer.
 *
 * \param dest       Pointer on the buffer to fill
 * \param nb_block   Number of block to transfer
 *
 * \return OK if started, otherwise FAIL
 */
uint32_t hsmci_start_read_blocks(void *dest, uint16_t nb_block);

/** \brief Wait the end of transfer initiated by mci_start_read_blocks()
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_wait_end_of_read_blocks(void);

/** \brief Start a write blocks transfer on the line
 * Note: The driver will use the DMA available to speed up the transfer.
 *
 * \param src        Pointer on the buffer to send
 * \param nb_block   Number of block to transfer
 *
 * \return OK if started, otherwise FAIL
 */
uint32_t hsmci_start_write_blocks(const void *src, uint16_t nb_block);

/** \brief Wait the end of transfer initiated by mci_start_write_blocks()
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t hsmci_wait_end_of_write_blocks(void);

#endif /* HSMCI_H_ */