/*
 * FreeRTOS+FAT SL V1.0.1 (C) 2014 HCC Embedded
 *
 * The FreeRTOS+FAT SL license terms are different to the FreeRTOS license
 * terms.
 *
 * FreeRTOS+FAT SL uses a dual license model that allows the software to be used
 * under a standard GPL open source license, or a commercial license.  The
 * standard GPL license (unlike the modified GPL license under which FreeRTOS
 * itself is distributed) requires that all software statically linked with
 * FreeRTOS+FAT SL is also distributed under the same GPL V2 license terms.
 * Details of both license options follow:
 *
 * - Open source licensing -
 * FreeRTOS+FAT SL is a free download and may be used, modified, evaluated and
 * distributed without charge provided the user adheres to version two of the
 * GNU General Public License (GPL) and does not remove the copyright notice or
 * this text.  The GPL V2 text is available on the gnu.org web site, and on the
 * following URL: http://www.FreeRTOS.org/gpl-2.0.txt.
 *
 * - Commercial licensing -
 * Businesses and individuals who for commercial or other reasons cannot comply
 * with the terms of the GPL V2 license must obtain a commercial license before
 * incorporating FreeRTOS+FAT SL into proprietary software for distribution in
 * any form.  Commercial licenses can be purchased from
 * http://shop.freertos.org/fat_sl and do not require any source files to be
 * changed.
 *
 * FreeRTOS+FAT SL is distributed in the hope that it will be useful.  You
 * cannot use FreeRTOS+FAT SL unless you agree that you use the software 'as
 * is'.  FreeRTOS+FAT SL is provided WITHOUT ANY WARRANTY; without even the
 * implied warranties of NON-INFRINGEMENT, MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. Real Time Engineers Ltd. and HCC Embedded disclaims all
 * conditions and terms, be they implied, expressed, or statutory.
 *
 * http://www.FreeRTOS.org
 * http://www.FreeRTOS.org/FreeRTOS-Plus
 *
 */
#include "api_mdriver.h"
#include "psp_string.h"
#include "HSMCI.h"
#include "sd_mmc_protocol.h"


/* The F_DRIVER structure that is filled with the SD versions of the read
sector, write sector, etc. functions. */
static F_DRIVER  s_driver;

/* Disk not initialized yet. */
static char in_use = 0;

//! Number of block to read or write on the current transfer
static uint16_t sd_mmc_nb_block_to_tranfer = 0;
//! Number of block remaining to read or write on the current transfer
static uint16_t sd_mmc_nb_block_remaining = 0;
//relative card address
uint16_t rca = 0;
//card specific data
uint8_t csd[CSD_REG_BSIZE];
//SD/MMC transfer rate unit codes (10K) list
const uint32_t sd_mmc_trans_units[7] = { 10, 100, 1000, 10000, 0, 0, 0 };
//SD transfer multiplier factor codes (1/10) list
const uint32_t sd_trans_multipliers[16] = {	0, 10, 12, 13, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 70, 80 };
//card capacity
uint32_t capacity;
//HSMCI clock and HS flag
uint32_t clock;
uint8_t high_speed;

uint32_t sd_cmd8(uint8_t * v2)
{
	uint32_t resp;

	*v2 = 0;
	// Test for SD version 2
	if (!hsmci_send_cmd(SD_CMD8_SEND_IF_COND, SD_CMD8_PATTERN | SD_CMD8_HIGH_VOLTAGE))
	{
		printf("SD card not v2\n\r");
		return OK; // It is not a V2
	}
	// Check R7 response
	resp = hsmci_get_response();
	if (resp == 0xFFFFFFFF)
	{
		// No compliance R7 value
		printf("SD card not v2\n\r");
		return OK; // It is not a V2
	}
	if ((resp & (SD_CMD8_MASK_PATTERN | SD_CMD8_MASK_VOLTAGE)) != (SD_CMD8_PATTERN | SD_CMD8_HIGH_VOLTAGE))
	{
		return FAIL;
	}
	printf("SD card v2\n\r");
	*v2 = 1;
	return OK;
}

/**
 * \brief Ask to all cards to send their operations conditions (MCI only).
 * - ACMD41 sends operation condition command.
 * - ACMD41 reads OCR
 *
 * \param v2   Shall be 1 if it is a SD card V2
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t sd_mci_op_cond(uint8_t v2)
{
	uint32_t arg, retry, resp;

	/*
	 * Timeout 1s = 400KHz / ((6+6+6+6)*8) cylces = 2100 retry
	 * 6 = cmd byte size
	 * 6 = response byte size
	 * 6 = cmd byte size
	 * 6 = response byte size
	 */
	retry = 2100;
	do
	{
		// CMD55 - Indicate to the card that the next command is an
		// application specific command rather than a standard command.
		if (!hsmci_send_cmd(SDMMC_CMD55_APP_CMD, 0))
		{
			return FAIL;
		}

		// (ACMD41) Sends host OCR register
		arg = SD_MMC_VOLTAGE_SUPPORT;
		if (v2)
		{
			arg |= SD_ACMD41_HCS;
		}
		// Check response
		if (!hsmci_send_cmd(SD_MCI_ACMD41_SD_SEND_OP_COND, arg))
		{
			return FAIL;
		}
		resp = hsmci_get_response();
		if (resp & OCR_POWER_UP_BUSY)
		{
			// Card is ready
			if ((resp & OCR_CCS) != 0)
			{
				//sd_mmc_card->type |= CARD_TYPE_HC;
				printf("Card is High Capacity\n\r");
			}
			break;
		}
		if (retry-- == 0)
		{
			return FAIL;
		}
	} while (1);
	return OK;
}

/**
 * \brief Decodes SD CSD register
 */
void sd_decode_csd(void)
{
 	uint32_t unit;
	uint32_t mul;
	uint32_t tran_speed;

	// Get SD memory maximum transfer speed in Hz.
	tran_speed = CSD_TRAN_SPEED(csd);
	unit = sd_mmc_trans_units[tran_speed & 0x7];
	mul = sd_trans_multipliers[(tran_speed >> 3) & 0xF];
	clock = unit * mul * 1000;

	/*
	 * Get card capacity.
	 * ----------------------------------------------------
	 * For normal SD/MMC card:
	 * memory capacity = BLOCKNR * BLOCK_LEN
	 * Where
	 * BLOCKNR = (C_SIZE+1) * MULT
	 * MULT = 2 ^ (C_SIZE_MULT+2)       (C_SIZE_MULT < 8)
	 * BLOCK_LEN = 2 ^ READ_BL_LEN      (READ_BL_LEN < 12)
	 * ----------------------------------------------------
	 * For high capacity SD card:
	 * memory capacity = (C_SIZE+1) * 512K byte
	 */
	if (CSD_STRUCTURE_VERSION(csd) >= SD_CSD_VER_2_0)
	{
		capacity =	(SD_CSD_2_0_C_SIZE(csd) + 1) * 512;
		printf("CSD version 2.0\n\r");
		printf("Card capacity is : %d KB\n\r",capacity);
	}
	else
	{
		uint32_t blocknr = ((SD_CSD_1_0_C_SIZE(csd) + 1) *	(1 << (SD_CSD_1_0_C_SIZE_MULT(csd) + 2)));
		capacity = blocknr * (1 << SD_CSD_1_0_READ_BL_LEN(csd)) / 1024;
	}
}

/**
 * \brief ACMD51 - Read the SD Configuration Register.
 *
 * \note
 * SD Card Configuration Register (SCR) provides information on the SD Memory
 * Card's special features that were configured into the given card. The size
 * of SCR register is 64 bits.
 *
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t sd_acmd51(void)
{
	uint8_t scr[SD_SCR_REG_BSIZE];
	
	// CMD55 - Indicate to the card that the next command is an
	// application specific command rather than a standard command.
	if (!hsmci_send_cmd(SDMMC_CMD55_APP_CMD, rca << 16))
	{
		return FAIL;
	}
	if (!hsmci_adtc_start(SD_ACMD51_SEND_SCR, 0, SD_SCR_REG_BSIZE, 1, OK))
	{
		return FAIL;
	}
	if (!hsmci_start_read_blocks(scr, 1))
	{
		return FAIL;
	}
	if (!hsmci_wait_end_of_read_blocks())
	{
		return FAIL;
	}

	// Get SD Memory Card - Spec. Version
	switch (SD_SCR_SD_SPEC(scr))
	{
		case SD_SCR_SD_SPEC_1_0_01:
		//sd_mmc_card->version = CARD_VER_SD_1_0;
		printf("SD spec. version 1.0\n\r");
		break;

		case SD_SCR_SD_SPEC_1_10:
		//sd_mmc_card->version = CARD_VER_SD_1_10;
		printf("SD spec. version 1.1\n\r");
		break;

		case SD_SCR_SD_SPEC_2_00:
		if (SD_SCR_SD_SPEC3(scr) == SD_SCR_SD_SPEC_3_00)
		{
			//sd_mmc_card->version = CARD_VER_SD_3_0;
			printf("SD spec. version 3.0\n\r");
		}
		else
		{
			//sd_mmc_card->version = CARD_VER_SD_2_0;
			printf("SD spec. version 2.0\n\r");
		}
		break;

		default:
		//sd_mmc_card->version = CARD_VER_SD_1_0;
		printf("SD spec. version 1.0\n\r");
		break;
	}
	return OK;
}

/**
 * \brief ACMD6 - Define the data bus width to 4 bits bus
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t sd_acmd6(void)
{
	// Define the data bus width to 4 bits bus
	// CMD55 - Indicate to the card that the next command is an
	// application specific command rather than a standard command.
	if (!hsmci_send_cmd(SDMMC_CMD55_APP_CMD, rca << 16))
	{
		return FAIL;
	}
	// 10b = 4 bits bus
	if (!hsmci_send_cmd(SD_ACMD6_SET_BUS_WIDTH, 0x2))
	{
		return FAIL;
	}
	else
	{
		// Set 4-bit bus width
		HSMCI->HSMCI_SDCR = HSMCI_SDCR_SDCSEL_SLOTA | HSMCI_SDCR_SDCBUS_4;
		printf("4-bit bus width enabled\n\r");
	}
	return OK;
}

/**
 * \brief CMD6 for SD - Switch card in high speed mode
 *
 * \note CMD6 for SD is valid under the "trans" state.
 * \note high_speed is updated.
 * \note clock is updated.
 *
 * \return OK if success, otherwise FAIL
 */
uint32_t sd_cm6_set_high_speed(void)
{
	uint8_t switch_status[SD_SW_STATUS_BSIZE];

	if (!hsmci_adtc_start(SD_CMD6_SWITCH_FUNC,
						  SD_CMD6_MODE_SWITCH
						  | SD_CMD6_GRP6_NO_INFLUENCE
						  | SD_CMD6_GRP5_NO_INFLUENCE
						  | SD_CMD6_GRP4_NO_INFLUENCE
						  | SD_CMD6_GRP3_NO_INFLUENCE
						  | SD_CMD6_GRP2_DEFAULT
						  | SD_CMD6_GRP1_HIGH_SPEED,
							SD_SW_STATUS_BSIZE, 1, OK))
	{
		printf("adtc command failed\n\r");
		return FAIL;
	}
	if (!hsmci_start_read_blocks(switch_status, 1))
	{
		printf("CMD6 start read blocks error\n\r");
		return FAIL;
	}
	if (!hsmci_wait_end_of_read_blocks())
	{
		printf("CMD6 wait end of read blocks error\n\r");
		return FAIL;
	}

	if (hsmci_get_response() & CARD_STATUS_SWITCH_ERROR)
	{
		printf("%s: CMD6 CARD_STATUS_SWITCH_ERROR\n\r", __func__);
		return FAIL;
	}
	if (SD_SW_STATUS_FUN_GRP1_RC(switch_status)	== SD_SW_STATUS_FUN_GRP_RC_ERROR)
	{
		// Not supported, it is not a protocol error
		//return OK;
		printf("SD_SW_STATUS_FUN_GRP_RC_ERROR\n\r");
	}
	if (SD_SW_STATUS_FUN_GRP1_BUSY(switch_status))
	{
		printf("%s: CMD6 SD_SW_STATUS_FUN_GRP1_BUSY\n\r", __func__);
		return FAIL;
	}
	// CMD6 function switching period is within 8 clocks
	// after the end bit of status data.
	hsmci_send_clock();
	high_speed = 1;
	clock *= 2;
	printf("Clock set after CMD6 : %d Hz\n\r", clock);
	return OK;
}

/**
 * \brief CMD13 - Addressed card sends its status register.
 * This function waits the clear of the busy flag
 *
 * \return true if success, otherwise false
 */
uint32_t sd_mmc_cmd13(void)
{
	uint32_t nec_timeout;

	/* Wait for data ready status.
	 * Nec timing: 0 to unlimited
	 * However a timeout is used.
	 * 200 000 * 8 cycles
	 */
	nec_timeout = 200000;
	do
	{
		if (!hsmci_send_cmd(SDMMC_MCI_CMD13_SEND_STATUS, rca << 16))
		{
			return FAIL;
		}
		// Check busy flag
		if (hsmci_get_response() & CARD_STATUS_READY_FOR_DATA)
		{
			break;
		}
		if (nec_timeout-- == 0)
		{
			printf("%s: CMD13 Busy timeout\n\r", __func__);
			return FAIL;
		}
	} while (1);

	return OK;
}

/****************************************************************************
 * Read one sector
 ***************************************************************************/
uint32_t sd_mmc_start_read_block(F_DRIVER * driver, void *dest, uint16_t nb_block)
{
	uint16_t s;
	uint32_t resp, sr;
	//maximum blocks in SD
	uint32_t maxsector = ((capacity * 1024) / SD_MMC_BLOCK_SIZE);
	
	/* Not used. */
	( void ) driver;
	 
	if (nb_block >= maxsector)
	{
		printf("Block to read >= Maxsector\n\r");
		return SD_READ_FAIL;
	}
	
	if( in_use == 0 )
	{
		printf("Device not initialized\n\r");
		return SD_READ_FAIL;
	}
	
	if (!sd_mmc_cmd13())
	{
		printf("CMD13 data ready error\n\r");
		return SD_READ_FAIL;
	}	
	
	/* Locate offset into flash for block. */
	s += nb_block;
	 
	HSMCI->HSMCI_MR = HSMCI_MR_PDCMODE | HSMCI_MR_WRPROOF | HSMCI_MR_RDPROOF;
	HSMCI->HSMCI_BLKR = HSMCI_BLKR_BLKLEN(SD_MMC_BLOCK_SIZE);
	// Configure PDC transfer
	HSMCI->HSMCI_RPR = (uint32_t) dest;
	HSMCI->HSMCI_RCR = ((SD_MMC_BLOCK_SIZE) / 4);
	// Start transfer
	HSMCI->HSMCI_PTCR = HSMCI_PTCR_RXTEN;
    // Block Address
	HSMCI->HSMCI_ARGR = s;
	
	HSMCI->HSMCI_CMDR   =   HSMCI_CMDR_CMDNB(SDMMC_CMD17_READ_SINGLE_BLOCK)
	|   HSMCI_CMDR_MAXLAT
	|   HSMCI_CMDR_RSPTYP_48_BIT
	|   HSMCI_CMDR_TRTYP_SINGLE
	|   HSMCI_CMDR_TRDIR_READ
	|   HSMCI_CMDR_TRCMD_START_DATA
	|	HSMCI_CMDR_SPCMD_STD;
	
	// Wait end of command
	do
	{
		sr = HSMCI->HSMCI_SR;
		if (HSMCI->HSMCI_CMDR & SDMMC_RESP_CRC)
		{
			if (sr & (HSMCI_SR_CSTOE | HSMCI_SR_RTOE | HSMCI_SR_RENDE | HSMCI_SR_RCRCE | HSMCI_SR_RDIRE | HSMCI_SR_RINDE))
			{
				hsmci_reset();
				printf("HSMCI CMDR CRC error \n\r");
				return SD_READ_FAIL;
			}
		}
	} while (!(sr & HSMCI_SR_CMDRDY));

	if (HSMCI->HSMCI_CMDR & SDMMC_RESP_BUSY)
	{
		if (!hsmci_wait_busy())
		{
			printf("HSMCI CMDR wait busy error \n\r");
			return SD_READ_FAIL;
		}
	}
	
	while(HSMCI_SR_ENDRX != (HSMCI->HSMCI_SR  & HSMCI_SR_ENDRX));
	
	return SD_READ_OK;
}
/****************************************************************************
 * Write one sector
 ***************************************************************************/
uint32_t sd_mmc_start_write_block(F_DRIVER * driver, void *src, uint16_t nb_block)
{
	uint16_t d;
	uint32_t resp, sr;
	//maximum blocks in SD
	uint32_t maxsector = ((capacity * 1024) / SD_MMC_BLOCK_SIZE);
	
	/* Not used. */
	( void ) driver;
	
	if (nb_block >= maxsector)
	{
		printf("Block to write >= Maxsector\n\r");
		return SD_WRITE_FAIL;
	}
	
	if( in_use == 0 )
	{
		printf("Device not initialized\n\r");
		return SD_WRITE_FAIL;
	}
	
	/* Locate offset into flash for block. */
	d += nb_block;
	
	HSMCI->HSMCI_MR = HSMCI_MR_PDCMODE | HSMCI_MR_WRPROOF | HSMCI_MR_RDPROOF;
	HSMCI->HSMCI_BLKR = HSMCI_BLKR_BLKLEN(SD_MMC_BLOCK_SIZE);
	// Configure PDC transfer
	HSMCI->HSMCI_TPR = (uint32_t) src;
	HSMCI->HSMCI_TCR = ((SD_MMC_BLOCK_SIZE) / 4);
	// Start transfer
	HSMCI->HSMCI_PTCR = HSMCI_PTCR_TXTEN;
	// Block Address
	HSMCI->HSMCI_ARGR = d;
	
	HSMCI->HSMCI_CMDR   =   HSMCI_CMDR_CMDNB(SDMMC_CMD24_WRITE_BLOCK)
	|   HSMCI_CMDR_MAXLAT
	|   HSMCI_CMDR_RSPTYP_48_BIT
	|   HSMCI_CMDR_TRTYP_SINGLE
	|   HSMCI_CMDR_TRDIR_WRITE
	|   HSMCI_CMDR_TRCMD_START_DATA
	|	HSMCI_CMDR_SPCMD_STD;
	
	// Wait end of command
	do
	{
		sr = HSMCI->HSMCI_SR;
		if (HSMCI->HSMCI_CMDR & SDMMC_RESP_CRC)
		{
			if (sr & (HSMCI_SR_CSTOE | HSMCI_SR_RTOE | HSMCI_SR_RENDE | HSMCI_SR_RCRCE | HSMCI_SR_RDIRE | HSMCI_SR_RINDE))
			{
				hsmci_reset();
				printf("HSMCI CMDR CRC error \n\r");
				return SD_WRITE_FAIL;
			}
		}
	} while (!(sr & HSMCI_SR_CMDRDY));

	if (HSMCI->HSMCI_CMDR & SDMMC_RESP_BUSY)
	{
		if (!hsmci_wait_busy())
		{
			printf("HSMCI CMDR wait busy error \n\r");
			return SD_WRITE_FAIL;
		}
	}
	
	while(HSMCI_SR_NOTBUSY != (HSMCI->HSMCI_SR  & HSMCI_SR_NOTBUSY));
	return SD_WRITE_OK;
}


/****************************************************************************
 *
 * sd_getphy
 *
 * determinate sd drive physicals
 *
 * INPUTS
 *
 * driver - driver structure
 * phy - this structure has to be filled with physical information
 *
 * RETURNS
 *
 * error code or zero if successful
 *
 ***************************************************************************/
static int sd_getphy ( F_DRIVER * driver, F_PHY * phy )
{
	/* Not used. */
	( void ) driver;

	phy->number_of_sectors = ((capacity * 1024)/SD_MMC_BLOCK_SIZE);
	phy->bytes_per_sector = SD_MMC_BLOCK_SIZE;

	return SD_READ_OK;
}

/****************************************************************************
 *
 * sd_release
 *
 * Releases a drive
 *
 * INPUTS
 *
 * driver_param - driver parameter
 *
 ***************************************************************************/
static void sd_release ( F_DRIVER * driver )
{
	/* Not used. */
	( void ) driver;

	/* Disk no longer in use. */
	in_use = 0;
}

/****************************************************************************
 *
 * sd_init
 *
 * this init function has to be passed for high level to initiate the
 * driver functions
 *
 * INPUTS
 *
 * driver_param - driver parameter
 *
 * RETURNS
 *
 * driver structure pointer
 *
 ***************************************************************************/
F_DRIVER * sd_init ( unsigned long driver_param )
{
	( void ) driver_param;

	if( in_use )
	return NULL;
		
	s_driver.readsector = sd_mmc_start_read_block;
	s_driver.writesector = sd_mmc_start_write_block;
	s_driver.getphy = sd_getphy;
	s_driver.release = sd_release;
	
	//SD card version 2
	uint8_t v2 = 0;
	
	// Card need of 74 cycles clock minimum to start
	hsmci_send_clock();
	
	// CMD0 - Reset all cards to idle state.
	if (!hsmci_send_cmd(SDMMC_MCI_CMD0_GO_IDLE_STATE, 0))
	{
		printf("CMD0 failed\n\r");
		return FAIL;
	}
	
	// Test SD card for voltage compatibility
	if (!sd_cmd8(&v2))
	{
		//return FAIL;
		printf("CMD8 failed\n\r");
	}
	
	// Try to get the SD card's operating condition
	if (!sd_mci_op_cond(v2))
	{
		//It is not a SD card
		//return FAIL;
		printf("Not a valid SD card\n\r");
	}
	
	// SD MEMORY, Put the Card in Identify Mode
	// Note: The CID is not used in this stack
	if (!hsmci_send_cmd(SDMMC_CMD2_ALL_SEND_CID, 0))
	{
		printf("CMD2 failed\n\r");
		return FAIL;
	}
	// Ask the card to publish a new relative address (RCA).
	if (!hsmci_send_cmd(SD_CMD3_SEND_RELATIVE_ADDR, 0))
	{
		printf("CMD3 failed\n\r");
		return FAIL;
	}
	rca = ((hsmci_get_response() >> 16) & 0xFFFF);
	printf("Card relative address is : %d\n\r", rca);
	
	// SD MEMORY, Get the Card-Specific Data
	if (!hsmci_send_cmd(SDMMC_MCI_CMD9_SEND_CSD, rca << 16))
	{
		printf("CMD9 failed\n\r");
		return FAIL;
	}
	hsmci_get_response_128(csd);
	sd_decode_csd();
	
	// Select the SD card and put it into Transfer Mode
	if (!hsmci_send_cmd(SDMMC_CMD7_SELECT_CARD_CMD, rca << 16))
	{
		printf("Failed to execute CMD7\n\r");
		return FAIL;
	}
	
	// SD MEMORY, Read the SCR to get card version
	if (!sd_acmd51())
	{
		//return FAIL;
		printf("ACMD51 failed\n\r");
	}
	
	// TRY to enable High-Speed Mode
	if (!sd_cm6_set_high_speed())
	{
		//return FAIL;
		printf("High Speed not set\n\r");
	}
	else
	{
		hsmci_select_device(0, clock, 1, high_speed);
		// Set power saving to maximum value
		//HSMCI->HSMCI_MR = HSMCI_MR_PWSDIV(0x07);
		//HSMCI->HSMCI_CR = HSMCI_CR_PWSEN;
		printf("Card switched to HS mode\n\r");
	}
	
	// Set bus width to 4-bit
	if (!sd_acmd6())
	{
		//return FAIL;
		printf("ACMD6 failed\n\r");
	}

	// SD MEMORY, Set default block size
	if (!hsmci_send_cmd(SDMMC_CMD16_SET_BLOCKLEN, SD_MMC_BLOCK_SIZE))
	{
		//return FAIL;
		printf("CMD16 failed\n\r");
	}
	printf("Card relative address is : %d\n\r", rca);

	in_use = 1;

	return &s_driver;
}



